from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from django.utils.encoding import force_text

from ..util import DMP_OPTIONS, merge_dicts

import os
import os.path



##################################################
###   Static File Provider Factory


DEFAULT_CONTENT_PROVIDERS = [
    { 'provider': 'django_mako_plus.CssLinkProvider' },
    { 'provider': 'django_mako_plus.JsLinkProvider'  },
    { 'provider': 'django_mako_plus.JsContextProvider' },
    # deprecated as of Oct 2017
    { 'provider': 'django_mako_plus.MakoCssProvider'  },
    { 'provider': 'django_mako_plus.MakoJsProvider'  },
]


def init_providers():
    '''Called when the DMP template engine is created by Django'''
    DMP_OPTIONS['RUNTIME_PROVIDERS'] = [ ProviderFactory(provider_def) for provider_def in DMP_OPTIONS.get('CONTENT_PROVIDERS', DEFAULT_CONTENT_PROVIDERS) ]
    DMP_OPTIONS['RUNTIME_PROVIDERS'].sort(key=lambda pf: pf.options['weight'], reverse=True)


class ProviderFactory(object):
    '''Creator for a given Provider definition in settings.py.'''
    def __init__(self, provider_def):
        self.options = {}
        try:
            self.provider_class = provider_def['provider']
        except KeyError:
            raise ImproperlyConfigured('The Django Mako Plus template OPTIONS were not set up correctly in settings.py; a CONTENT_PROVIDERS item is missing `provider`.')
        if isinstance(self.provider_class, str):
            self.provider_class = import_string(self.provider_class)
        if not issubclass(self.provider_class, BaseProvider):
            raise ImproperlyConfigured('The Django Mako Plus template OPTIONS were not set up correctly in settings.py; The `provider` value must be a subclass of django_mako_plus.BaseProvider.')
        self.options = merge_dicts(self.provider_class.default_options, provider_def)
        
    def create(self, app_dir, template_name, version_id):
        return self.provider_class(app_dir, template_name, self.options, version_id)
        
        

##############################################################
###   Static File Providers

class BaseProvider(object):
    '''
    A list of providers is attached to each template as it is process.
    These are cached with the template, so .__init__() is only called once
    per system runtime, while .append_content() is called for each
    request.  Optimize the methods accordingly.
    '''
    default_options = {
        'group': 'styles',
        'weight': 0,  # higher weights run first
    }
    def __init__(self, app_dir, template_name, options, version_id):
        self.app_name = os.path.split(app_dir)[1]                  # django app name
        self.app_dir = app_dir                                     # absolute path
        self.template_name = os.path.splitext(template_name)[0]    # without the extension
        self.options = merge_dicts(self.default_options, options)  # combined options dictionary
        self.version_id = version_id                               # unique number for overriding the cache (see LinkProvider)
        self.init()
        
        
    @property
    def group(self):
        return self.options['group']


    def init(self):
        '''Called at the end of the constructor.'''
        
        
    def get_content(self, provider_run):
        '''
        Called each time the template renders.  Should return a string 
        to be included in the HTML output, or None if no content.
        '''
        return None
        
        
    def format_string(self, val):
        '''
        Helper function that runs st.format with some standard named options.
        val.format() is called with {appname}, {appdir}, {template}.
        '''
        return force_text(val).format(
            appname=self.app_name,
            appdir=self.app_dir,
            template=self.template_name,
        )
        


from django.apps import apps
from django.template import Context

import mako.runtime

from ..util import get_dmp_instance
from ..uid import wuid
from .templateinfo import build_templateinfo_chain

import io
import warnings




############################################################
###  Public items in the package

from .base import init_providers, BaseProvider
from .compile import CompileProvider, CompileScssProvider, CompileLessProvider
from .links import LinkProvider, CssLinkProvider, JsLinkProvider, JsContextProvider, jscontext
from .mako_static import MakoCssProvider, MakoJsProvider

#########################################################
###  Primary functions

def links(tself, version_id=None, group=None):
    '''
    Returns the HTML for the given provider group.
    
        Add this at the top (<head> section) of your template:
        ${ django_mako_plus.static_group(self, 'styles') }
    
        Add this at the bottom of your template:
        ${ django_mako_plus.static_group(self, 'scripts') }
    
        Or, to get all content in one call:
        ${ django_mako_plus.static_group(self) }

    Suppose you have two html files connected via template inheritance: index.html
    and base.htm.  This method will render the links for all static files
    with the names of either of these two templates.  Assuming the default
    providers are listed in settings.py, up to four files are linked/sourced:

        base.htm       Generates a <style> link for app/styles/base.css
            |          and a <style> source for app/styles/base.cssm
            |
        index.html     Generates a <style> link for app/styles/index.css
                       and a <style> source for app/styles/index.cssm

    This call must be made from within a rendering Mako template because
    this is where you have access to the "self" namespace.
    If you need to render these links outside of a template, see link_template_css()
    below.

    The optional version_id parameter is to overcome browser caches.  On some browsers,
    changes to your CSS/JS files don't get downloaded because the browser waits a time
    to check for a new version.  This wait time is set by your web server,
    and it's normally a good thing to speed everything up.  However,
    when you upload new CSS/JS files, you want all browsers to download the new
    files even if their cached versions have't expired yet.
    By adding an arbitrary id to the end of the .css and .js files, browsers will
    see the files as *new* anytime that id changes.  The default method
    for calculating the id is the file modification time (minutes since 1970).
    '''
    request = tself.context.get('request')
    provider_run = ProviderRun(request, tself.context, group, build_templateinfo_chain(tself, version_id))
    return provider_run.get_content()


class ProviderRun(object):
    '''Information for a run through a chain of template info objects and providers on each one.'''
    def __init__(self, request, context, group, chain):
        self.uid = wuid()       # a unique Id for this run
        self.request = request  # request from the render call
        self.context = context  # context from the render call
        self.group = group      # the provider group being rendered (usually None)
        self.chain = chain      # list of TemplateInfos matching the template inheritance
        self.chain_index = 0    # current index of the inheritance chain as the run goes
        
    def get_content(self):
        '''Loops each TemplateInfo providers list, returning the combined content.'''
        self.chain_index = 0
        self.html = []
        for ti in self.chain:
            for provider_i, provider in enumerate(ti.providers):
                if self.group is None or provider.group == self.group:
                    content = provider.get_content(self)
                    if content:
                        self.html.append(content)
            self.chain_index += 1
        return '\n'.join(self.html)



def template_links(request, app, template_name, context=None, version_id=None, group=None, force=True):
    '''
    Returns the HTML for the given provider group, using an app and template name.
    This method should not normally be used (use links() instead).  The use of 
    this method is when provider need to be called from regular python code instead
    of from within a rendering template environment.
    
    See links() for documentation on the variables.
    '''
    if isinstance(app, str):
        app = apps.get_app_config(app)
    if context is None:
        context = {}

    # get the template object normally
    template = get_dmp_instance().get_template_loader(app, create=True).get_mako_template(template_name, force=force)
        
    # create a mako context so it seems like we are inside a render
    # I'm hacking into private Mako methods here, but I can't see another
    # way to do this.  Hopefully this can be rectified at some point.
    context_dict = {}
    if isinstance(context, Context):
        for d in context:
            context_dict.update(d)
    elif context is not None:
        context_dict.update(context)
    context_dict.pop('self', None)  # some contexts have self in them, and it messes up render_unicode below because we get two selfs
    runtime_context = mako.runtime.Context(io.BytesIO(), **context_dict)
    runtime_context._set_with_template(template)
    _, mako_context = mako.runtime._populate_self_namespace(runtime_context, template)
    return links(mako_context['self'], version_id=version_id, group=group)




#######################################################################
###   Deprecated methods - these are deprecated as of Oct 2017

def link_css(tself, version_id=None):
    '''
    Deprecated as of Oct 2017.
    Use `django_mako_plus.links(self, 'styles')` instead.
    '''
    warnings.warn("link_css() is deprecated as of Oct 2017.  Use `django_mako_plus.links(self, 'styles')` instead.", DeprecationWarning)
    return links(tself, version_id=version_id, group='styles')


def link_js(tself, version_id=None):
    '''
    Deprecated as of Oct 2017.
    Use `django_mako_plus.links(self, 'scripts')` instead.
    '''
    warnings.warn("link_js() is deprecated as of Oct 2017.  Use `django_mako_plus.links(self, 'scripts')` instead.", DeprecationWarning)
    return links(tself, version_id=version_id, group='scripts')


def link_template_css(request, app, template_name, context, version_id=None, force=True):
    '''
    Deprecated as of Oct 2017.
    Use `django_mako_plus.template_links(..., group='styles')` instead.
    '''
    warnings.warn("link_template_css() is deprecated as of Oct 2017.  Use `django_mako_plus.template_links(..., group='styles')` instead.", DeprecationWarning)
    return template_links(request, app, template_name, context, version_id=version_id, group='styles', force=force)


def link_template_js(request, app, template_name, context, version_id=None, force=True):
    '''
    Deprecated as of Oct 2017.
    Use `django_mako_plus.template_links(..., group='scripts')` instead.
    '''
    warnings.warn("link_template_js() is deprecated as of Oct 2017.  Use `django_mako_plus.template_links(..., group='scripts')` instead.", DeprecationWarning)
    return template_links(request, app, template_name, context, version_id=version_id, group='scripts', force=force)





from django.conf import settings
from django.utils.module_loading import import_string

from ..uid import wuid
from ..util import merge_dicts
from ..version import __version__
from .base import BaseProvider

import os
import os.path
import posixpath
import json


class LinkProvider(BaseProvider):
    '''Base class for providers that create links'''
    # the default options are for CSS files
    default_options = merge_dicts(BaseProvider.default_options, {  
        'group': 'styles',
        'filename': '{appdir}/somedir/{template}.static.file',
    })
    def init(self):
        self.href = None
        fullpath = self.format_string(self.options['filename'])
        if os.path.exists(fullpath):
            # the version_id is a unique number either given by the project or by reading the last modified time of the file
            # it is important to add to links (see make_link in the subclasses) because it makes the url unique, which
            # forces browsers to re-download the file despite a previous version being in their cache.  Without this id,
            # some browsers (Chrome, I'm looking at you) use the cached version even after updates to the file.
            if self.version_id is None:
                self.version_id = int(os.stat(fullpath).st_mtime)
            # make it relative to the project root, and ensure we have forward slashes (even on windows) because this is for urls
            self.href = posixpath.join(settings.STATIC_URL, os.path.relpath(fullpath, settings.BASE_DIR))  

    def get_content(self, provider_run):
        '''Subclasses should override this method'''
        return '<div class="DMP_LinkProvider">{}</div>'.format(self.href)


class CssLinkProvider(LinkProvider):
    '''Generates a CSS <link>'''
    default_options = merge_dicts(LinkProvider.default_options, { 
        'filename': '{appdir}/styles/{template}.css',
        'skip_duplicates': True,
    })
    def get_content(self, provider_run):
        if self.href is None:
            return None
        return '<link id="{uid}" data-context="{contextid}" rel="stylesheet" type="text/css" href="{href}?{version}" />'.format(
            uid=wuid(),           
            contextid=provider_run.uid,
            href=self.href, 
            version=self.version_id,
            skip_duplicates='true' if self.options['skip_duplicates'] else 'false',
        )


class JsLinkProvider(LinkProvider):
    '''Generates a JS <script>.'''
    default_options = merge_dicts(LinkProvider.default_options, {  
        'group': 'scripts',
        'filename': '{appdir}/scripts/{template}.js',
        'async': False,
    })
    def get_content(self, provider_run):
        if self.href is None:
            return None
        return '<script>DMP_CONTEXT.addScript("{uid}", "{contextid}", "{app}/{template}", "{href}?{version}", {async});</script>'.format(
            uid=wuid(),           
            contextid=provider_run.uid,  
            app=self.app_name.replace('"', '\\"'),
            template=self.template_name.replace('"', '\\"'),
            href=self.href, 
            version=self.version_id,
            async='true' if self.options['async'] else 'false',
        )


class JsContextProvider(BaseProvider):
    '''Adds all js_context() variables to DMP_CONTEXT'''
    default_options = merge_dicts(LinkProvider.default_options, {  
        'group': 'scripts',
        'encoder': 'django.core.serializers.json.DjangoJSONEncoder',
    })
    
    def init(self):
        super().init()
        self.encoder = import_string(self.options['encoder'])

    def get_content(self, provider_run):
        # send the context with the main template (would be redundant to do for ancestors)
        html = []
        html.append('<script>')
        if provider_run.chain_index == 0:
            context_data = { k: provider_run.context[k] for k in provider_run.context.kwargs if isinstance(k, jscontext) }
            html.append('DMP_CONTEXT.set("{version}", "{contextid}", {data});'.format(
                version=__version__,
                contextid=provider_run.uid,  
                data=json.dumps(context_data, cls=self.encoder, separators=(',', ':')) if context_data else '{}',
            ))
        html.append('DMP_CONTEXT.linkContextByName("{contextid}", "{app}/{template}");'.format(
            contextid=provider_run.uid,  
            app=self.app_name.replace('"', '\\"'),
            template=self.template_name.replace('"', '\\"'),
        ))
        html.append('</script>')
        return ('\n' if settings.DEBUG else '').join(html)



class jscontext(str):
    '''
    Marks a key in the context dictionary as a JS context item.
    JS context items are sent to the template like normal,
    but they are also added to the runtime JS namespace.
    
    See the tutorial for more information on this function.
    '''
    # no code needed, just using the class for identity
    



from django.core.exceptions import ImproperlyConfigured
from django.template import TemplateDoesNotExist

from ..util import get_dmp_instance, merge_dicts
from .base import BaseProvider

import os
import os.path

try:
    from rjsmin import jsmin
except ImportError:
    jsmin = None
try:
    from rcssmin import cssmin
except ImportError:
    cssmin = None



###  DEPRECATED as of DMP 4.3.1
class MakoCssProvider(BaseProvider):
    '''Provides the content for *.cssm files'''
    default_options = merge_dicts(BaseProvider.default_options, {  
        'group': 'styles',
        'minify': True,
    })
    def init(self):
        self.cssm_dir = os.path.join(self.app_dir, 'styles')
        try:
            self.template = get_dmp_instance().get_template_loader_for_path(self.cssm_dir).get_template(self.template_name + '.cssm')
        except TemplateDoesNotExist:
            self.template = None
        
    def get_content(self, provider_run):
        if self.template is not None:
            content = self.template.render(request=provider_run.request, context=provider_run.context)
            if self.options['minify']:
                if cssmin is not None:
                    content = cssmin(content)
                else:
                    raise ImproperlyConfigured("Unable to minify {}.cssm because rcssmin is not available. Please `pip install rcssmin`.".format(self.template_name))
            return '<style type="text/css">{}</style>'.format(content)
        return None
        

###  DEPRECATED as of DMP 4.3.1
class MakoJsProvider(BaseProvider):
    '''Provides the content for *.jsm files'''
    default_options = merge_dicts(BaseProvider.default_options, {  
        'group': 'scripts',
        'minify': True,
    })
    def init(self):
        self.cssm_dir = os.path.join(self.app_dir, 'scripts')
        try:
            self.template = get_dmp_instance().get_template_loader_for_path(self.cssm_dir).get_template(self.template_name + '.jsm')
        except TemplateDoesNotExist:
            self.template = None
        
    def get_content(self, provider_run):
        if self.template is not None:
            content = self.template.render(request=provider_run.request, context=provider_run.context)
            if self.options['minify']:
                if jsmin is not None:
                    content = jsmin(content)
                else:
                    raise ImproperlyConfigured("Unable to minify {}.jsm because rjsmin is not available. Please `pip install rjsmin`.".format(self.template_name))
            return '<script type="text/javascript">{}</script>'.format(content)
        return None





from ..command import run_command
from ..util import merge_dicts
from .base import BaseProvider

import os
import os.path
import shutil



class CompileProvider(BaseProvider):
    '''
    Compiles static files, such as *.scss or *.less, when an output file 
    timestamp is older than the source file. In production mode, this check
    is done only once (the first time a template is run) per server start.
    
    Special format keywords for use in the options:
        {appdir} - The app directory for the template being rendered (full path).
        {template} - The name of the template being rendered, without its extension.
    '''
    # the command line should be specified as a list (see the subprocess module)
    default_options = merge_dicts(BaseProvider.default_options, {  
        'group': 'styles',
        'weight': 10,  
        'source': '{appdir}/somedir/{template}.source',
        'output': '{appdir}/somedir/{template}.output',
        'command': [ 'echo', '{appdir}/somedir/{template}.source', '{appdir}/somedir/{template}.output' ],
    })
    def init(self):
        # doing the check in init() so it only happens one time during production
        source_path = self.format_string(self.options['source'])
        if os.path.exists(source_path):
            source_stat = os.stat(source_path)
            compiled_path = self.format_string(self.options['output'])
            compiled_exists = os.path.exists(compiled_path)
            compiled_stat = os.stat(compiled_path) if compiled_exists else None
            if not compiled_exists or source_stat.st_mtime > compiled_stat.st_mtime:
                run_command(*[ self.format_string(a) for a in self.options['command'] ])


class CompileScssProvider(CompileProvider):
    '''Specialized CompileProvider that contains settings for *.scss files.'''
    default_options = merge_dicts(CompileProvider.default_options, {  
        'source': '{appdir}/styles/{template}.scss',
        'output': '{appdir}/styles/{template}.css',
        'command': [ shutil.which('scss'), '--unix-newlines', '{appdir}/styles/{template}.scss', '{appdir}/styles/{template}.css' ],
    })


class CompileLessProvider(CompileProvider):
    '''Specialized CompileProvider that contains settings for *.less files.'''
    default_options = merge_dicts(CompileProvider.default_options, {  
        'source': '{appdir}/styles/{template}.less',
        'output': '{appdir}/styles/{template}.css',
        'command': [ shutil.which('lessc'), '--source-map', '{appdir}/styles/{template}.less', '{appdir}/styles/{template}.css' ],
    })

    


from django.conf import settings

from ..util import DMP_OPTIONS

import os
import os.path

# key to attach TemplateInfo objects and static renderer to the Mako Template itself
# I attach it to the template objects because they are already cached by mako, so
# caching them again here would result in double-caching.  It's a bit of a
# monkey-patch to set them as attributes in the Mako templates, but it's efficient
# and encapsulated entirely within this file
DMP_TEMPLATEINFO_KEY = '_django_mako_plus_templateinfo'



###################################################
###  TemplateInfo: attached to each template

class TemplateInfo(object):
    '''
    Data class that holds information about a template's directories.  A TemplateInfo object
    is created by _build_templateinfo_chain for each level in the Mako template inheritance chain.
    This object is then attached to the template object, which Mako already caches.
    That way we only do this work once per server run (in production mode).
    '''
    def __init__(self, app_dir, template_name, version_id):
        self.app_dir = app_dir              # absolute path
        self.template_name = template_name  # without the extension
        self.providers = [ pf.create(app_dir, template_name, version_id) for pf in DMP_OPTIONS['RUNTIME_PROVIDERS'] ]


def build_templateinfo_chain(tself, version_id):
    '''
    Retrieves a chain of TemplateInfo objects.  The chain is formed by following
    template inheritance through inherit tags: <%inherit file="parent.htm" />.
    '''
    # step through the template inheritance
    chain = []
    while tself is not None:
        # first check the cache, creating if necessary
        ti = getattr(tself.template, DMP_TEMPLATEINFO_KEY, None)
        if ti is None:
            template_dir, template_name = os.path.split(tself.template.filename)
            app_dir = os.path.dirname(template_dir)
            ti = TemplateInfo(app_dir, template_name, version_id)
            # cache in template if in production mode
            if not settings.DEBUG:
                setattr(tself.template, DMP_TEMPLATEINFO_KEY, ti)
        # append the TemplateInfo
        chain.append(ti)
        # loop with the next inherited template
        tself = tself.inherits

    # return
    chain.reverse() # we need furthest ancestor first, current template last
    return chain

    
    



from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

import base64


# this is populated with the dictionary of options in engine.py when
# Django initializes the template engine
DMP_OPTIONS = {}

# the key for the DMP singleton engine instance (stored in DMP_OPTIONS by engine.py)
DMP_INSTANCE_KEY = 'django_mako_plus_instance'

# set up the logger
import logging
log = logging.getLogger('django_mako_plus')

################################################################
###   Utility functions

def get_dmp_instance():
    '''
    Retrieves the DMP template engine instance.
    '''
    try:
        return DMP_OPTIONS[DMP_INSTANCE_KEY]
    except KeyError:
        raise ImproperlyConfigured('The Django Mako Plus template engine did not initialize correctly.  Check your logs for previous errors that may have caused initialization to fail, and check that DMP is set correctly in settings.py.')


def get_dmp_app_configs():
    '''
    Gets the DMP-enabled app configs, which will be a subset of all installed apps.  This is a generator function.
    This method does not use the registry cache.  Instead, it looks through all Django apps in settings.py and
    yields any with "DJANGO_MAKO_PLUS = True" in its __init__.py.
    '''
    for config in apps.get_app_configs():
        # check for the DJANGO_MAKO_PLUS = True in the app (should be in app/__init__.py)
        if getattr(config.module, 'DJANGO_MAKO_PLUS', False):
            yield config



################################################################
###   Special type of list used for url params

class URLParamList(list):
    '''
    A simple extension to Python's list that returns '' for indices that don't exist.
    For example, if the object is ['a', 'b'] and you call obj[5], it will return ''
    rather than throwing an IndexError.  This makes dealing with url parameters
    simpler since you don't have to check the length of the list.
    '''
    def __getitem__(self, idx):
        '''Returns the element at idx, or '' if idx is beyond the length of the list'''
        return self.get(idx, '')

    def get(self, idx, default=''):
        '''Returns the element at idx, or default if idx is beyond the length of the list'''
        # if the index is beyond the length of the list, return ''
        if isinstance(idx, int) and (idx >= len(self) or idx < -1 * len(self)):
            return default
        # else do the regular list function (for int, slice types, etc.)
        return super().__getitem__(idx)


##########################################
###   Utilities

def merge_dicts(*dicts):
    '''
    Merges an arbitrary number of dicts, starting 
    with the first argument and updating through the
    last argument (last dict wins on conflicting keys).
    '''
    merged = {}
    for d in dicts:
        merged.update(d)
    return merged
    
 




from mako.lexer import Lexer
from mako import parsetree

from .exceptions import SassCompileException
from .util import DMP_OPTIONS
from .command import run_command

import glob, sys, os, os.path, subprocess, itertools, re
from warnings import warn

#    ___             _ __                                    _                 _
#   |   \    ___    | '_ \    _ _    ___     __     __ _    | |_     ___    __| |
#   | |) |  / -_)   | .__/   | '_|  / -_)   / _|   / _` |   |  _|   / -_)  / _` |
#   |___/   \___|   |_|__   _|_|_   \___|   \__|_  \__,_|   _\__|   \___|  \__,_|
# _|"""""|_|"""""|_|"""""|_|"""""|_|"""""|_|"""""|_|"""""|_|"""""|_|"""""|_|"""""|
# "`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'
#

__doc__ = '''
DEPRECATED: instead use CompileProvider (see the docs).

Compiles scss files using the Sass runtime arguments in settings.py.

Also compiles scssm files, which are scss files with embedded Mako.
Only ${ var } expressions are supported right now.  Other Mako
constructs like <% %> may be supported later.
'''


MAKO_EXPRESSION_MARKERS = ( 'MakoExpression', 'ExpressionMako' )
RE_MAKO_EXPRESSION = re.compile('{}_(.+?)_{}'.format(MAKO_EXPRESSION_MARKERS[0], MAKO_EXPRESSION_MARKERS[1]), flags=re.MULTILINE)


def check_template_scss(styles_dir, template_name):
    '''
    Checks to see if template.scss or template.scssm need compiling.
    The styles_dir should be the full path to the styles directory.
    The template_name should be the name of the template *without the extension*.

    Raises a SassCompileException if compilation fails.
    '''
    warn("The django_mako_plus.scss is deprecated and will be removed at some point. Please switch to CompileProvider instead.")
    for ext in ( 'css', 'cssm' ):
        scss_file = os.path.join(styles_dir, '{}.s{}'.format(template_name, ext))
        gen_css_file = os.path.join(styles_dir, '{}.{}'.format(template_name, ext))
        try:
            scss_stat = os.stat(scss_file)
        except OSError:
            scss_stat = None
        if scss_stat is not None:  # only continue this block if we found a .scss file
            try:
                fstat = os.stat(gen_css_file)
            except OSError:
                fstat = None
            # if we 1) have no css_file or 2) have a newer scss_file, run the compiler
            if fstat is None or scss_stat.st_mtime > fstat.st_mtime:
                try:
                    if ext == 'css':
                        compile_scss_file(scss_file, gen_css_file)
                    else:
                        compile_scssm_file(scss_file, gen_css_file)
                except subprocess.CalledProcessError as cpe:
                    raise SassCompileException(cpe.cmd, cpe.stderr)


def compile_scss_file(scss_file, css_file):
    '''
    Compiles a regular scss file

    If the Sass execution fails, this function raises subprocess.CalledProcessError.
    '''
    # run sass on it
    args = list(DMP_OPTIONS.get('RUNTIME_SCSS_ARGUMENTS'))
    args.append(scss_file)
    args.append(css_file)
    run_command(*args)


def compile_scssm_file(scssm_file, css_file):
    '''
    Compiles an scssm (sass with embedded Mako) file.
    This is a total hack to sneak mako codes through the Sass compiler.  The algorithm
    uses the Mako lexer to split the file, then it replaces Expressions in the file
    with Base32 -- that alphabet is valid throughout CSS rules.  After Sass does its thing,
    we switch the Base32 back to the Mako codes.

    If the Sass execution fails, this function raises subprocess.CalledProcessError.
    '''
    # this algorithm isn't terribly fast, but it only runs the first time the file is accessed on a production system
    # lock the file so other processes can't access it
    encoding = settings.DEFAULT_CHARSET or 'utf8'
    with lock_file(scssm_file):

        # read the contents of the file
        with open(scssm_file, encoding=encoding) as fin:
            contents = fin.read()

        try:
            # make a backup just in case something goes wrong (this is for testing)
            # backup_scss = '{}.bak'.format(scssm_file)
            # with open(backup_scss, 'w', encoding=encoding) as fout:
            #     fout.write(contents)
            # embed mako codes in markers that are acceptable to sass
            with open(scssm_file, 'w', encoding=encoding) as fout:
                for node in Lexer(text=contents).parse().nodes:
                    if isinstance(node, parsetree.Expression):
                        fout.write('{}_{}_{}'.format(MAKO_EXPRESSION_MARKERS[0], encode32(node.text), MAKO_EXPRESSION_MARKERS[1]))
                    elif isinstance(node, parsetree.Text):
                        fout.write(node.content)
            # run sass on it
            compile_scss_file(scssm_file, css_file)
            # read the contents of the generated file
            with open(css_file, encoding=encoding) as fin:
                csscontents = fin.read()
            # unencode all the mako expressions
            csscontents = RE_MAKO_EXPRESSION.sub(lambda match: '${{{}}}'.format(decode32(match.group(1))), csscontents)
            # write the contents
            with open(css_file, 'w', encoding=encoding) as fout:
                fout.write(csscontents)
            # erase the backup
            # os.remove(backup_scss)

        finally:
            # replace the original contents back in the scss file
            with open(scssm_file, 'w', encoding=encoding) as fout:
                fout.write(contents)
            # update the modified timestamp on the css file so the scssm_file is older than it (we just rewrote the scssm)
            os.utime(css_file)



#################################################################
###   File locking context manager - used by sass.py

class lock_file(object):
    '''
    A context manager that provides a non-blocking file lock on both Unix and Windows.
    Uses a temporary file for the lock - in the same directory as the filename.
    If the lock fails after timeout_seconds, an OSError is raised.

    I'm not using fcntl.flock because it is Unix-only and DMP is used on both
    Unix and Windows.  This is a very simplistic locking scheme, but it works
    for our purposes.
    '''
    def __init__(self, filename, timeout_seconds=5):
        self.filename = filename
        self.lock_filename = '{}.templock'.format(filename)
        self.timeout_seconds = timeout_seconds


    def __enter__(self):
        # acquire the lock to it
        for i in range(self.timeout_seconds):
            # create the lock file
            if not os.path.exists(self.lock_filename):
                self.fd = open(self.lock_filename, 'w')
                # we created successfully, so break out!
                break
            else:  # couldn't get the lock, so wait a second a try again
                time.sleep(1)
        else:  # we couldn't get a lock in timeout_seconds tries, so raise the exception
            raise OSError("Unable to acquire lock on {}; if you are sure no other processes are running, you may need to delete the lock file manually: {}".format(self.filename, self.lock_filename))
        # return
        return self


    def __exit__(self, exec_type, exec_val, exec_tb):
        # close and remove the temporary file
        self.fd.close()
        try:
            os.remove(self.lock_filename)
        except FileNotFoundError:  # shouldn't ever happen, but just in case
            pass





######################################################
###   Encoding routines.
###   Using Base32 because its alphabet conforms to
###   CSS selectors.

def encode32(st):
    '''Encodes the given string to base64.'''
    if not isinstance(st, bytes):
        st = st.encode('utf8')              # we now have a byte string rather than the original unicode text
    b32_byte_st = base64.b32encode(st)   # we now have a base32-encoded byte string representing the text
    return b32_byte_st.decode('ascii').replace('=', '9')    # we're now back to Unicode (using ascii decoding since base32 is all ascii characters)


def decode32(st):
    '''Decodes the given base64-encoded string.'''
    st = st.replace('9', '=')
    if not isinstance(st, bytes):
        st = st.encode('ascii')     # we now have a byte string of the base64-encoded text instead of the Unicode base32s-encoded st
    byte_st = base64.b32decode(st)   # we now have a byte string of the original text
    return byte_st.decode('utf8')         # we now have a Unicode string of the original text


